==> quantagonia-0.5.0-py3-none-any.whl: quantagonia/mip/model.py <==
from __future__ import annotations

import io
import os
import sys
import tempfile

from pyscipopt import Expr as Expression
from pyscipopt import Model as _Model

from quantagonia import HybridSolver
from quantagonia.enums import HybridSolverOptSenses, HybridSolverStatus, VarType
from quantagonia.errors.errors import ModelError
from quantagonia.extras import SuppressScipOutput
from quantagonia.mip.variable import Variable


class Model:
    """A class representing a Mixed Integer Programming (MIP) problem instance.

    Args:
        name (str): (optional) The name of the MIP model. Defaults to "model".

    """

    def __init__(self, name: str = "model"):
        """Initializes the Model instance with a given name.

        Args:
            name (str): (optional) The name of the model. Defaults to "model".

        Raises:
            ValueError: If the QUANTAGONIA_API_KEY environment variable is not set.
        """
        api_key = os.getenv("QUANTAGONIA_API_KEY")
        if api_key is None:
            msg = "QUANTAGONIA_API_KEY not found"
            raise ValueError(msg)
        self._hybrid_solver = HybridSolver(api_key)

        self._model = _Model(name)
        self._model.redirectOutput()  # Redirect output to stdout
        self._sol_status = HybridSolverStatus.UNKNOWN
        self._best_bound = None
        self._solution = None
        self._best_objective = None
        self._best_time = None
        self._total_time = None

    @property
    def sense(self) -> HybridSolverOptSenses:
        """The optimization sense of the model.

        Returns:
            HybridSolverOptSenses: The current optimization sense of the model.

        Raises:
            ModelError: If the current sense is unsupported.
        """
        if self._model.getObjectiveSense() == "minimize":
            return HybridSolverOptSenses.MINIMIZE
        if self._model.getObjectiveSense() == "maximize":
            return HybridSolverOptSenses.MAXIMIZE
        error_message = "Unsupported sense."
        raise ModelError(error_message)

    @property
    def objective_value(self) -> float:
        """The objective value of the model's best solution.

        Returns:
            float: The objective value of the best solution found.
        """
        return self._best_objective

    @property
    def solution_status(self) -> HybridSolverStatus:
        """The solution status of the model.

        Returns:
            HybridSolverStatus: The current solution status of the model.
        """
        return self._sol_status

    @property
    def solution(self) -> dict[str, float]:
        """The solution of the model.

        Returns:
            dict[str, float]: A dictionary mapping variable names to their values in the solution.
        """
        return self._solution

    @property
    def best_bound(self) -> float:
        """The best bound of the model.

        Returns:
            float: The best bound found during optimization.
        """
        return self._best_bound

    @property
    def best_time(self) -> float:
        """The runtime (in seconds) after which the best solution was found.

        Returns:
            float: The runtime (in seconds) after which the best solution was found.
        """
        return self._best_time

    @property
    def total_time(self) -> float:
        """The total runtime of the optimization process.

        Returns:
            float: The total runtime (in seconds) of the optimization process.
        """
        return self._total_time

    @property
    def objective_offset(self) -> float:
        """The offset of the objective function.

        Returns:
            float: The constant offset value of the objective function.
        """
        return self._model.getObjoffset()

    def add_variable(
        self,
        lb: float = 0,
        ub: float = float("Inf"),
        name: str = "",
        coeff: float = 0,
        var_type: VarType = VarType.CONTINUOUS,
    ) -> Variable:
        """Adds a new variable to the model.

        Args:
            lb (float): The lower bound of the variable. Defaults to 0.
            ub (float): The upper bound of the variable. Defaults to positive infinity.
            name (str): The name of the variable. Defaults to an empty string.
            coeff (float): The coefficient of the variable in the objective function. Defaults to 0.
            var_type (VarType): The type of the variable. Defaults to VarType.CONTINUOUS.

        Returns:
            Variable: The Variable instance that was added to the MIP model.

        Raises:
            ModelError: If the variable type is unsupported or if the variable addition fails.
        """
        if var_type == VarType.CONTINUOUS:
            vtype = "C"
        elif var_type == VarType.BINARY:
            vtype = "B"
        elif var_type == VarType.INTEGER:
            vtype = "I"
        else:
            error_message = "Unsupported variable type."
            raise ModelError(error_message)
        try:
            # Suppress SCIP output
            with SuppressScipOutput():
                var = self._model.addVar(name, vtype, lb, ub, coeff)
        except ValueError as e:
            error_message = f"Failed to add variable '{name}'."
            raise ModelError(error_message) from e
        return Variable(var, self._model)

    def add_constraint(self, expr: Expression, name: str = "") -> None:
        """Adds a constraint to the model.

        Args:
            expr (Expression): The expression representing the constraint.
            name (str): (optional) The name of the constraint.

        Raises:
            ModelError: If the constraint addition fails.
        """
        try:
            # Suppress SCIP output
            with SuppressScipOutput():
                self._model.addCons(expr, name)
        except ValueError as e:
            error_message = f"Failed to add constraint '{name}'."
            raise ModelError(error_message) from e

    def add_constraints(self, constraints: list[tuple[Expression, str]]) -> None:
        """Adds multiple constraints to the model.

        Args:
            constraints (List[Tuple[Expression, str]]): A list of tuples, each containing
                an expression and a name for a constraint.
        """
        for constraint in constraints:
            try:
                # Suppress SCIP output
                with SuppressScipOutput():
                    self.add_constraint(constraint[0], constraint[1])
            except ValueError as e:
                error_message = f"Failed to add constraint '{constraint[1]}'."
                raise ModelError(error_message) from e

    def set_objective(self, expr: Expression, sense: HybridSolverOptSenses = HybridSolverOptSenses.MINIMIZE) -> None:
        """Sets the objective function for the model.

        Args:
            expr (Expression): The expression representing the objective function.
            sense (HybridSolverOptSenses): The optimization sense (minimize or maximize).
                Defaults to HybridSolverOptSenses.MINIMIZE.

        Raises:
            ModelError: If the sense is unsupported or if setting the objective fails.
        """
        if sense == HybridSolverOptSenses.MINIMIZE:
            ref_sense = "minimize"
        elif sense == HybridSolverOptSenses.MAXIMIZE:
            ref_sense = "maximize"
        else:
            error_message = "Unsupported sense."
            raise ModelError(error_message)
        try:
            # Suppress SCIP output
            with SuppressScipOutput():
                self._model.setObjective(expr=expr, sense=ref_sense, clear=False)
        except ValueError as e:
            error_message = "Failed to set objective."
            raise ModelError(error_message) from e

    def add_objective_offset(self, offset: float) -> None:
        """Adds an offset to the objective function of the model.

        Args:
            offset (float): The offset value to be added to the objective function.
        """
        self._model.addObjoffset(offset)

    def set_sense(self, sense: HybridSolverOptSenses) -> None:
        """Sets the optimization sense of the model.

        Args:
            sense (HybridSolverOptSenses): The optimization sense to be set.

        Raises:
            ModelError: If the provided sense is unsupported.
        """
        if sense == HybridSolverOptSenses.MINIMIZE:
            self._model.setMinimize()
        elif sense == HybridSolverOptSenses.MAXIMIZE:
            self._model.setMaximize()
        else:
            error_message = "Unsupported sense."
            raise ModelError(error_message)

    def _assign_solution(self, solution: dict[str, float]) -> None:
        """Assigns the solution to the reference model.

        Args:
            solution (dict[str, float]): A dictionary mapping variable names to their values.

        Raises:
            RuntimeError: If the solution has an incorrect length or cannot be assigned to a variable.
            ValueError: If the cloud solution is not feasible for the local model.
            ModelError: If the cloud solution cannot be added to the local model.
        """
        variables = self._model.getVars()
        if len(solution) != len(variables):
            msg = "Solution has incorrect length"
            raise RuntimeError(msg)

        # create local solution
        try:
            # Suppress SCIP output
            with SuppressScipOutput():
                local_solution = self._model.createOrigSol()
        except ValueError as e:
            msg = "Model failed to assign the solution computed by HybridSolver"
            raise RuntimeError(msg) from e

        # inject cloud solution to local solution
        for var in variables:
            try:
                # Suppress SCIP output
                with SuppressScipOutput():
                    self._model.setSolVal(local_solution, var, solution[str(var)])
            except Exception as e:
                msg = f"Couldn't assign solution to variable {var}"
                raise RuntimeError(msg) from e

        # check if HybridSolver solution is feasible
        feasible = self._model.checkSol(local_solution)
        if not feasible:
            msg = "Cloud solution is not feasible for local model"
            raise ValueError(msg)

        # add cloud solution to local model
        added = self._model.addSol(local_solution)
        if not added:
            msg = "Couldn't add cloud solution to local model"
            raise ModelError(msg)

    def _write_problem(self, problem_path: str) -> None:
        # When redirectOutput() is used, writing the model prints the model to stdout instead of the file.
        # This is a known issue in pyscipopt: https://github.com/scipopt/PySCIPOpt/issues/243.
        # For now, we capture the output and write it to a file.

        # Redirect stdout to a StringIO object
        captured_output = io.StringIO()
        sys.stdout = captured_output
        # Write the problem, which will now be captured
        self._model.writeProblem(problem_path, verbose=False)
        # Restore stdout
        sys.stdout = sys.__stdout__
        # Remove comment lines starting with '*' from the captured output
        filter_comments = [line for line in captured_output.getvalue().splitlines() if not line.startswith("*")]
        filtered_output = "\n".join(filter_comments)
        # Write the captured output to a file
        with open(problem_path, "w") as f:
            f.write(filtered_output)

    def optimize(self) -> HybridSolverStatus:
        """Sends the model to the HybridSolver and solves it on the cloud.

        Returns:
            HybridSolverStatus: The solution status after optimization.

        Raises:
            ValueError: If the solution status cannot be cast to HybridSolverStatus.
        """
        # using the context manager to ensure the tmp_path is deleted even if an error occurs
        with tempfile.TemporaryDirectory() as tmp_path:
            # the https client is taking care of gzipping before sending the file, so no gzipping here
            tmp_problem = os.path.join(tmp_path, self._model.getProbName() + ".mps")
            try:
                self._write_problem(tmp_problem)
                res, _ = self._hybrid_solver.solve(tmp_problem)
            except ValueError as e:
                msg = "Failed to submit the problem to the HybridSolver"
                raise ModelError(msg) from e

        # assign the solution to the local model
        self._solution = res["solution"]
        # assign the solution to the _model; disabled for now
        # self._assign_solution(self._solution)  # noqa: ERA001

        self._best_objective = res["objective"]
        self._best_bound = res["bound"]
        self._best_time = res["best_time"]
        self._total_time = res["timing"]

        # assign solution status
        try:
            self._sol_status = HybridSolverStatus(res["sol_status"])
        except ValueError as exc:
            msg = f"Couldn't cast solution status {res['sol_status']}"
            raise ValueError(msg) from exc

        return self._sol_status

==> quantagonia-0.7.0-py3-none-any.whl: quantagonia/mip/model.py <==
from __future__ import annotations

import io
import os
import sys
import tempfile

from pyscipopt import Expr as Expression
from pyscipopt import Model as _Model

from quantagonia import HybridSolver
from quantagonia.enums import HybridSolverOptSenses, HybridSolverStatus, VarType
from quantagonia.mip.wrappers.wrapper_enums import GRB, CplexEnum, PythonMipEnum
from quantagonia.errors.errors import ModelError
from quantagonia.extras import SuppressScipOutput
from quantagonia.mip.variable import Variable


class Model:
    """A class representing a Mixed Integer Programming (MIP) problem instance.

    Args:
        name (str): (optional) The name of the MIP model. Defaults to "model".

    """

    def __init__(self, name: str = "Model") -> None:
        """Initializes the Model instance with a given name.

        Args:
            name (str): (optional) The name of the model. Defaults to "Model".

        Raises:
            ValueError: If the QUANTAGONIA_API_KEY environment variable is not set.
        """
        api_key = os.getenv("QUANTAGONIA_API_KEY")
        if api_key is None:
            msg = "QUANTAGONIA_API_KEY not found"
            raise ValueError(msg)
        self._hybrid_solver = HybridSolver(api_key)

        self._name = name
        self._model = _Model(name)
        self._model.redirectOutput()  # Redirect output to stdout
        self._sol_status = HybridSolverStatus.UNKNOWN
        self._best_bound = None
        self._solution = None
        self._best_objective = None
        self._best_time = None
        self._total_time = None
        self._objective_function = None
        self._type = "BaseModel"

    @property
    def type(self) -> str:
        """The type of the model.

        Returns:
            str: The type of the model.
        """
        return self._type

    @property
    def model(self) -> _Model:
        return self._model

    @property
    def name(self) -> str:
        """The name of the model.

        Returns:
            str: The name of the model.
        """
        return self._name

    @property
    def sense(self) -> HybridSolverOptSenses:
        """The optimization sense of the model.

        Returns:
            HybridSolverOptSenses: The current optimization sense of the model.

        Raises:
            ModelError: If the current sense is unsupported.
        """
        if self._model.getObjectiveSense() == "minimize":
            return HybridSolverOptSenses.MINIMIZE
        if self._model.getObjectiveSense() == "maximize":
            return HybridSolverOptSenses.MAXIMIZE
        error_message = "Unsupported sense."
        raise ModelError(error_message)

    @sense.setter
    def sense(self, sense: HybridSolverOptSenses | GRB | CplexEnum | PythonMipEnum | str) -> None:
        """Sets the optimization sense of the model.

        Args:
            sense (HybridSolverOptSenses): The optimization sense to be set.

        Raises:
            ModelError: If the provided sense is unsupported.
        """
        self.set_sense(sense)

    @property
    def objective_value(self) -> float:
        """The objective value of the model's best solution.

        Returns:
            float: The objective value of the best solution found.
        """
        return self._best_objective

    @property
    def solution_status(self) -> HybridSolverStatus:
        """The solution status of the model.

        Returns:
            HybridSolverStatus: The current solution status of the model.
        """
        return self._sol_status

    @property
    def solve_status(self) -> HybridSolverStatus:
        """The solution status of the model.

        Returns:
            HybridSolverStatus: The current solution status of the model.
        """
        return self.solution_status

    # python-mip interface
    @property
    def status(self) -> HybridSolverStatus:
        """The solution status of the model.

        Returns:
            HybridSolverStatus: The current solution status of the model.
        """
        return self.solution_status

    @property
    def solution(self) -> dict[str, float]:
        """The solution of the model.

        Returns:
            dict[str, float]: A dictionary mapping variable names to their values in the solution.
        """
        return self._solution

    @solution.setter
    def solution(self, solution: dict[str, float]) -> None:
        """Sets the solution of the model.

        Args:
            solution (dict[str, float]): A dictionary mapping variable names to their values in the solution.
        """
        self._solution = solution

    # python-mip interface
    @property
    def num_solutions(self) -> int:
        """The number of solutions found during optimization.

        Returns:
            int: The number of solutions found during the MIP search.
        """
        return 1 if self._solution is not None else 0

    @property
    def best_bound(self) -> float:
        """The best bound of the model.

        Returns:
            float: The best bound found during optimization.
        """
        return self._best_bound

    @property
    def best_time(self) -> float:
        """The runtime (in seconds) after which the best solution was found.

        Returns:
            float: The runtime (in seconds) after which the best solution was found.
        """
        return self._best_time

    @property
    def total_time(self) -> float:
        """The total runtime of the optimization process.

        Returns:
            float: The total runtime (in seconds) of the optimization process.
        """
        return self._total_time

    @property
    def objective_offset(self) -> float:
        """The offset of the objective function.

        Returns:
            float: The constant offset value of the objective function.
        """
        return self._model.getObjoffset()

    @property
    def n_variables(self) -> int:
        """The number of variables in the model.

        Returns:
            int: The number of variables in the model.
        """
        return self._model.getNVars()

    @property
    def n_bin_vars(self) -> int:
        """The number of binary variables in the model.

        Returns:
            int: The number of binary variables in the model.
        """
        return self._model.getNBinVars()

    @property
    def n_int_vars(self) -> int:
        """The number of integer variables in the model.

        Returns:
            int: The number of integer variables in the model.
        """
        return self._model.getNIntVars()

    @property
    def n_cont_vars(self) -> int:
        """The number of continuous variables in the model.

        Returns:
            int: The number of continuous variables in the model.
        """
        return self.n_variables - self.n_bin_vars - self.n_int_vars

    def add_variable(
        self,
        lb: float = 0,
        ub: float = float("Inf"),
        name: str = "",
        coeff: float = 0,
        var_type: VarType | GRB | CplexEnum | PythonMipEnum = VarType.CONTINUOUS,
    ) -> Variable:
        """Adds a new variable to the model.

        Args:
            lb (float): The lower bound of the variable. Defaults to 0.
            ub (float): The upper bound of the variable. Defaults to positive infinity.
            name (str): The name of the variable. Defaults to an empty string.
            coeff (float): The coefficient of the variable in the objective function. Defaults to 0.
            var_type (VarType): The type of the variable. Defaults to VarType.CONTINUOUS.

        Returns:
            Variable: The Variable instance that was added to the MIP model.

        Raises:
            ModelError: If the variable type is unsupported or if the variable addition fails.
        """
        if var_type in [VarType.CONTINUOUS, GRB.CONTINUOUS, CplexEnum.CONTINUOUS, PythonMipEnum.CONTINUOUS]:
            vtype = "C"
        elif var_type in [VarType.BINARY, GRB.BINARY, CplexEnum.BINARY, PythonMipEnum.BINARY]:
            vtype = "B"
        elif var_type in [VarType.INTEGER, GRB.INTEGER, CplexEnum.INTEGER, PythonMipEnum.INTEGER]:
            vtype = "I"
        else:
            error_message = "Unsupported variable type."
            raise ModelError(error_message)
        try:
            # Suppress SCIP output
            with SuppressScipOutput():
                var = self._model.addVar(name, vtype, lb, ub, coeff)
        except ValueError as e:
            error_message = f"Failed to add variable '{name}'."
            raise ModelError(error_message) from e
        return Variable(var, self)

    def get_variables(self) -> list[Variable]:
        """Returns a list of all variables in the model.

        Returns:
            list[Variable]: A list of all variables in the model.
        """
        return [Variable(var, self) for var in self._model.getVars()]

    def add_constraint(self, expr: Expression, name: str = "") -> None:
        """Adds a constraint to the model.

        Args:
            expr (Expression): The expression representing the constraint.
            name (str): (optional) The name of the constraint.

        Raises:
            ModelError: If the constraint addition fails.
        """
        try:
            # Suppress SCIP output
            with SuppressScipOutput():
                self._model.addCons(expr, name)
        except ValueError as e:
            error_message = f"Failed to add constraint '{name}'."
            raise ModelError(error_message) from e

    def add_constraints(self, constraints: list[tuple[Expression, str]]) -> None:
        """Adds multiple constraints to the model.

        Args:
            constraints (List[Tuple[Expression, str]]): A list of tuples, each containing
                an expression and a name for a constraint.
        """
        for constraint in constraints:
            try:
                # Suppress SCIP output
                with SuppressScipOutput():
                    self.add_constraint(constraint[0], constraint[1])
            except ValueError as e:
                error_message = f"Failed to add constraint '{constraint[1]}'."
                raise ModelError(error_message) from e

    def set_objective(
        self,
        sense: HybridSolverOptSenses | GRB | CplexEnum | PythonMipEnum | str = HybridSolverOptSenses.MINIMIZE,
        expr: Expression = None,
    ) -> None:
        """Sets the objective function for the model.

        Args:
            expr (Expression): The expression representing the objective function.
            sense (HybridSolverOptSenses): The optimization sense (minimize or maximize).
                Defaults to HybridSolverOptSenses.MINIMIZE.

        Raises:
            ModelError: If the sense is unsupported or if setting the objective fails.
        """
        if expr is None:
            error_message = "Objective expression is missing."
            raise ModelError(error_message)
        if sense in [HybridSolverOptSenses.MINIMIZE, GRB.MINIMIZE, CplexEnum.Minimize, PythonMipEnum.MINIMIZE, "min"]:
            ref_sense = "minimize"
        elif sense in [HybridSolverOptSenses.MAXIMIZE, GRB.MAXIMIZE, CplexEnum.Maximize, PythonMipEnum.MAXIMIZE, "max"]:
            ref_sense = "maximize"
        else:
            error_message = "Unsupported sense."
            raise ModelError(error_message)
        try:
            # Suppress SCIP output
            with SuppressScipOutput():
                self._objective_function = expr
                self._model.setObjective(expr=expr, sense=ref_sense, clear=False)
        except ValueError as e:
            error_message = "Failed to set objective."
            raise ModelError(error_message) from e

    def add_objective_offset(self, offset: float) -> None:
        """Adds an offset to the objective function of the model.

        Args:
            offset (float): The offset value to be added to the objective function.
        """
        self._model.addObjoffset(offset)

    def set_sense(self, sense: HybridSolverOptSenses | GRB | CplexEnum | PythonMipEnum | str) -> None:
        """Sets the optimization sense of the model.

        Args:
            sense (HybridSolverOptSenses): The optimization sense to be set.

        Raises:
            ModelError: If the provided sense is unsupported.
        """
        if sense in [HybridSolverOptSenses.MINIMIZE, GRB.MINIMIZE, CplexEnum.Minimize, PythonMipEnum.MINIMIZE, "min"]:
            self._model.setMinimize()
        elif sense in [HybridSolverOptSenses.MAXIMIZE, GRB.MAXIMIZE, CplexEnum.Maximize, PythonMipEnum.MAXIMIZE, "max"]:
            self._model.setMaximize()
        else:
            error_message = "Unsupported sense."
            raise ModelError(error_message)

    def _assign_solution(self, solution: dict[str, float]) -> None:
        """Assigns the solution to the reference model.

        Args:
            solution (dict[str, float]): A dictionary mapping variable names to their values.

        Raises:
            RuntimeError: If the solution has an incorrect length or cannot be assigned to a variable.
            ValueError: If the cloud solution is not feasible for the local model.
            ModelError: If the cloud solution cannot be added to the local model.
        """
        variables = self._model.getVars()
        if len(solution) != len(variables):
            msg = "Solution has incorrect length"
            raise RuntimeError(msg)

        # create local solution
        try:
            # Suppress SCIP output
            with SuppressScipOutput():
                local_solution = self._model.createOrigSol()
        except ValueError as e:
            msg = "Model failed to assign the solution computed by HybridSolver"
            raise RuntimeError(msg) from e

        # inject cloud solution to local solution
        for var in variables:
            try:
                # Suppress SCIP output
                with SuppressScipOutput():
                    self._model.setSolVal(local_solution, var, solution[str(var)])
            except Exception as e:
                msg = f"Couldn't assign solution to variable {var}"
                raise RuntimeError(msg) from e

        # check if HybridSolver solution is feasible
        feasible = self._model.checkSol(local_solution)
        if not feasible:
            msg = "Cloud solution is not feasible for local model"
            raise ValueError(msg)

        # add cloud solution to local model
        added = self._model.addSol(local_solution)
        if not added:
            msg = "Couldn't add cloud solution to local model"
            raise ModelError(msg)

    def _write_problem(self, problem_path: str) -> None:
        # When redirectOutput() is used, writing the model prints the model to stdout instead of the file.
        # This is a known issue in pyscipopt: https://github.com/scipopt/PySCIPOpt/issues/243.
        # For now, we capture the output and write it to a file.

        # Redirect stdout to a StringIO object
        captured_output = io.StringIO()
        sys.stdout = captured_output
        # Write the problem, which will now be captured
        self._model.writeProblem(problem_path, verbose=False)
        # Restore stdout
        sys.stdout = sys.__stdout__
        # Remove comment lines starting with '*' from the captured output
        filter_comments = [line for line in captured_output.getvalue().splitlines() if not line.startswith("*")]
        filtered_output = "\n".join(filter_comments)
        # Write the captured output to a file
        with open(problem_path, "w") as f:
            f.write(filtered_output)

    def _optimize(self) -> HybridSolverStatus:
        """Optimizes the model using the HybridSolver.

        Returns:
            HybridSolverStatus: The solution status after optimization.

        Raises:
            ModelError: If the problem cannot be written to a file or if the solution
            status cannot be cast to HybridSolverStatus.
        """
        # using the context manager to ensure the tmp_path is deleted even if an error occurs
        with tempfile.TemporaryDirectory() as tmp_path:
            # the https client is taking care of gzipping before sending the file, so no gzipping here
            tmp_problem = os.path.join(tmp_path, self._model.getProbName() + ".mps")
            try:
                self._write_problem(tmp_problem)
                res, _ = self._hybrid_solver.solve(tmp_problem)
            except ValueError as e:
                msg = "Failed to submit the problem to the HybridSolver"
                raise ModelError(msg) from e

        # assign the solution to the local model
        self._solution = res["solution"]
        # assign the solution to the _model; disabled for now
        # self._assign_solution(self._solution)  # noqa: ERA001

        self._best_objective = res["objective"]
        self._best_bound = res["bound"]
        self._best_time = res["best_time"]
        self._total_time = res["timing"]

        # assign solution status
        try:
            self._sol_status = HybridSolverStatus(res["sol_status"])
        except ValueError as exc:
            msg = f"Couldn't cast solution status {res['sol_status']}"
            raise ValueError(msg) from exc

        return self._sol_status

    def optimize(self) -> HybridSolverStatus:
        """Sends the model to the HybridSolver and solves it on the cloud.

        Returns:
            HybridSolverStatus: The solution status after optimization.

        Raises:
            ValueError: If the solution status cannot be cast to HybridSolverStatus.
        """
        return self._optimize()
